/* FILE: eulerevolve.cc                 -*-Mode: c++-*-
 *
 * Concrete evolver class, using simple forward Euler steps
 *
 */

#include <float.h>

#include "nb.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "eulerevolve.h"
#include "key.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_EulerEvolve);

/* End includes */

// Constructor
Oxs_EulerEvolve::Oxs_EulerEvolve(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Evolver(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  const vector<string>*  vsptr; // Temp pointer

  gamma=2.21e5;  // Default value
  vsptr = FindInitValue("gamma");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("gamma",1);
    gamma=Nb_Atof((*vsptr)[0].c_str());
    DeleteInitValue("gamma");
  }

  // alpha is required input
  CheckInitValueParamCount("alpha",1);
  alpha=Nb_Atof((*FindInitValue("alpha"))[0].c_str());
  DeleteInitValue("alpha");

  do_precess=1; // Default value
  vsptr = FindInitValue("do_precess");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("do_precess",1);
    do_precess=atoi((*vsptr)[0].c_str());
    DeleteInitValue("do_precess");
  }

  start_dm=0.01; // Default value
  vsptr = FindInitValue("start_dm");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("start_dm",1);
    start_dm=Nb_Atof((*vsptr)[0].c_str());
    DeleteInitValue("start_dm");
  }

  VerifyAllInitArgsUsed();
}

BOOL Oxs_EulerEvolve::Init()
{
  next_timestep=0.0;
  state_id=0;
  max_dm_dt=-1; // Dummy value
  dE_dt=0;      // Dummy value
  return 1;
}

Oxs_EulerEvolve::~Oxs_EulerEvolve()
{}



void Oxs_EulerEvolve::Calculate_dm_dt
(const Oxs_Mesh& mesh_,
 const Oxs_MeshValue<REAL8m>& Ms_,
 const Oxs_MeshValue<ThreeVector>& mxH_,
 const Oxs_MeshValue<ThreeVector>& spin_,
 Oxs_MeshValue<ThreeVector>& dm_dt_,
 REAL8m& max_dm_dt_,REAL8m& dE_dt_) const
{ // Imports: mesh_, Ms_, mxH_, spin_
  // Exports: dm_dt_, max_dm_dt_, dE_dt_
  UINT4m size = mesh_.Size();  // Assume all imports are compatible
  REAL8m coef1 = -fabs(gamma);
  REAL8m coef2 = -fabs(alpha);
  ThreeVector scratch;
  REAL8m max_dm_dt_sq=0.0;
  REAL8m dm_dt_sq;
  REAL8m dE_dt_sum=0.0;
  for(UINT4m i=0;i<size;i++) {
    if(Ms_[i]==0) {
      dm_dt_[i].Set(0.0,0.0,0.0);
    } else {
      scratch = mxH_[i];
      dE_dt_sum += scratch.MagSq() * Ms_[i] * mesh_.Volume(i);
      scratch *= coef1;
      if(do_precess) {
	dm_dt_[i] = scratch;
      } else {
	dm_dt_[i].Set(0.0,0.0,0.0);
      }
      scratch ^= spin_[i]; // -|gamma|((mxH)xm)
      scratch *= coef2;  // |alpha.gamma|((mxH)xm) = -|alpha.gamma|(mx(mxH))
      dm_dt_[i] += scratch;
      dm_dt_sq = dm_dt_[i].MagSq();
      if(dm_dt_sq>max_dm_dt_sq) max_dm_dt_sq=dm_dt_sq;
    }
  }
  max_dm_dt_ = sqrt(max_dm_dt_sq);
  dE_dt_ = -1 * MU0 * fabs(gamma*alpha) * dE_dt_sum;
  return;
}




BOOL
Oxs_EulerEvolve::Step(const Oxs_Driver& driver,
		      Oxs_ConstKey<Oxs_SimState> current_state,
		      Oxs_Key<Oxs_SimState>& next_state)
{
  UINT4m size,i; // Mesh size and indexing variable

  const Oxs_SimState& cstate = current_state.GetReadReference();
  Oxs_SimState& nstate = next_state.GetWriteReference();
  nstate.valid_state=0;  // Safety

  if(cstate.mesh->Id() != nstate.mesh->Id()) {
    throw Oxs_Ext::Error
      ("EulerEvolve::Step: Oxs_Mesh not fixed across steps.");
  }

  REAL8m stepsize = next_timestep;
  if( cstate.Id() != state_id || stepsize<=0.0) {
    energy.AdjustSize(cstate.mesh);
    mxH.AdjustSize(cstate.mesh);
    dm_dt.AdjustSize(cstate.mesh);
    director->GetTotalEnergyAndmxH(cstate,energy,mxH);
    Calculate_dm_dt(*(cstate.mesh),*(cstate.Ms),
		    mxH,cstate.spin,
		    dm_dt,max_dm_dt,dE_dt);
    if(start_dm < sqrt(DBL_MAX/4) * max_dm_dt) {
      stepsize = start_dm / max_dm_dt;
    } else {
      stepsize = sqrt(DBL_MAX/4);
    }
    state_id = cstate.Id();
  }

  if (!cstate.valid_state) {
    // cstate is invalid, but presumably it is valid after
    // this routine fills in what it knows.  So copy cstate
    // to nstate, make what corrections we can, mark nstate
    // valid and return.
    nstate = cstate;
    nstate.last_timestep=0;
    nstate.iteration_count++;
    nstate.stage_iteration_count++;
    nstate.max_dm_dt = max_dm_dt;
    REAL8m total_energy=0.0;
    size = cstate.mesh->Size();
    for(i=0;i<size;++i) {
      total_energy += energy[i] * cstate.mesh->Volume(i);
    }
    nstate.total_energy = total_energy;
    nstate.valid_state=1;
    next_state.GetReadReference();  // Release write lock
    state_id = nstate.Id();
    next_timestep = stepsize;
    return 1;
  }

  // Negotiate with driver over size of next step
  nstate.last_timestep=stepsize;
  driver.FillStatePartB(nstate);
  BOOL forcestep=0;
  if(nstate.last_timestep>stepsize) {
    // driver wants to force this stepsize
    forcestep=1;
  }
  stepsize = nstate.last_timestep;

  nstate.elapsed_time = cstate.elapsed_time + stepsize;
  nstate.iteration_count = cstate.iteration_count + 1;
  nstate.stage_iteration_count = cstate.stage_iteration_count + 1;

  // Put new spin configuration in next_state
  size = nstate.spin.Size();
  ThreeVector tempspin;
  for(i=0;i<size;++i) {
    tempspin = dm_dt[i];
    tempspin *= stepsize;
    tempspin += cstate.spin[i];
    tempspin.SetMag(1.0);
    nstate.spin[i] = tempspin;
  }

  //  Calculate delta E
  new_energy.AdjustSize(nstate.mesh);
  mxH.AdjustSize(nstate.mesh);
  dm_dt.AdjustSize(nstate.mesh);
  director->GetTotalEnergyAndmxH(nstate,new_energy,mxH);
  REAL8m dE=0.0;
  for(i=0;i<size;++i) {
    dE += (new_energy[i] - energy[i]) * nstate.mesh->Volume(i);
  }

  // Check step and adjust next_timestep
  if(!forcestep && dE>0.0) {
    nstate.valid_state=0;  // Safety
    next_state.GetReadReference();  // Release write lock
    next_timestep=stepsize/2.0;
    return 0;  // Reject step
  }
  REAL8m expected_dE = dE_dt * stepsize;
  // NOTE: Both dE and expected_dE are negative
  if(dE > 0.5*expected_dE) next_timestep = 0.5*stepsize;
  else {
    next_timestep = stepsize * 1.2;
    if(dE > expected_dE) next_timestep *= dE/expected_dE;
    /// Because both dE's are <0, the ratio dE/expected_dE
    /// in the previous line is <1, i.e., dE>expected_dE
    /// means we got less of a drop than expected.
  }

  // Otherwise, accept step
  Calculate_dm_dt(*(nstate.mesh),
		  *(nstate.Ms),
		  mxH,nstate.spin,
		  dm_dt,max_dm_dt,dE_dt);
  nstate.max_dm_dt = max_dm_dt;
  nstate.total_energy = cstate.total_energy + dE;
  nstate.valid_state=1;
  next_state.GetReadReference();  // Release write lock
  energy.Swap(new_energy);
  state_id = nstate.Id();

  return 1;  // Good step
}
