/* FILE: exchange6ngbr.cc            -*-Mode: c++-*-
 *
 * 6 neighbor exchange energy on rectangular mesh,
 * derived from Oxs_Energy class.
 *
 */

#include "nb.h"
#include "key.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "exchange6ngbr.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_Exchange6Ngbr);

/* End includes */


// Constructor
Oxs_Exchange6Ngbr::Oxs_Exchange6Ngbr(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("A",1);
  A=Nb_Atof((*FindInitValue("A"))[0].c_str());
  DeleteInitValue("A");

  CheckInitValueParamCount("mesh",1);
  const string& meshname = (*FindInitValue("mesh"))[0];
  Oxs_Ext* tempobj = director->FindExtObject(meshname);
  if(tempobj==NULL) {
    string msg=string("Unable to identifu Oxs_Mesh object with name ")
      + meshname;
    throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_RectangularMesh* mesh
    = dynamic_cast<Oxs_RectangularMesh*>(tempobj);
  if(mesh==NULL) {
    string msg="Object " + meshname + " is not a rectangular mesh.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  meshkey.Set(mesh); // Dependency lock is held until *this is deleted.
  DeleteInitValue("mesh");

  VerifyAllInitArgsUsed();
}

Oxs_Exchange6Ngbr::~Oxs_Exchange6Ngbr()
{}

BOOL Oxs_Exchange6Ngbr::Init()
{
  return 1;
}

void Oxs_Exchange6Ngbr::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  // Check that the mesh hasn't changed.
  if(meshkey.KeyId() != state.mesh->Id()) {
    string msg="Input mesh to Oxs_Exchange6Ngbr::GetEnergyAndField()"
      " routine of object " +  string(InstanceName())
      + " is different than the one used to initialize this object.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  const Oxs_RectangularMesh* mesh = meshkey.GetPtr();
  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  REAL8m wgtx = -A/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = -A/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = -A/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  REAL8m hcoef = -2/MU0;

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address
	ThreeVector base = spin[i];
	REAL8m Msii = Ms_inverse[i];
	if(Msii == 0.0) {
	  energy[i]=0.0;
	  field[i].Set(0.,0.,0.);
	  continue;
	}
	ThreeVector sum(0.,0.,0.);
	if(z>0) {
	  UINT4m j = i-xydim;
	  if(Ms_inverse[j]!=0.0) sum += wgtz*(spin[j] - base);
	}
	if(y>0) {
	  UINT4m j = i-xdim;
	  if(Ms_inverse[j]!=0.0) sum += wgty*(spin[j] - base);
	}
	if(x>0) {
	  UINT4m j = i-1;
	  if(Ms_inverse[j]!=0.0) sum += wgtx*(spin[j] - base);
	}
	if(x<xdim-1) {
	  UINT4m j = i+1;
	  if(Ms_inverse[j]!=0.0) sum += wgtx*(spin[j] - base);
	}
	if(y<ydim-1) {
	  UINT4m j = i+xdim;
	  if(Ms_inverse[j]!=0.0) sum += wgty*(spin[j] - base);
	}
	if(z<zdim-1) {
	  UINT4m j = i+xydim;
	  if(Ms_inverse[j]!=0.0) sum += wgtz*(spin[j]- base);
	}
	field[i] = (hcoef*Msii) * sum;
	energy[i] = (sum * base);
      }
    }
  }
}
