/* FILE: fft.cc             -*-Mode: c++-*-
 *
 * C++ code to do 1 and 2 dimensional FFT's.
 *
 */

#include "nb.h"  // For constants PI and SQRT1_2
#include "oxsexcept.h"
#include "fft.h"

/* End includes */

#undef USE_COMPLEX

#define CMULT(xr,xi,yr,yi,zr,zi) (zr) = (xr)*(yr)-(xi)*(yi), \
                                 (zi) = (xr)*(yi)+(xi)*(yr)

////////////////////////////////////////////////////////////////////////
// 1D Complex FFT
void FFT::ReleaseMemory(void) const
{
  if(vecsize>0) {
    if(Uforward!=(MyComplex *)NULL)  delete[] Uforward;
    if(Uinverse!=(MyComplex *)NULL)  delete[] Uinverse;
    if(workspace!=(MyComplex *)NULL) delete[] workspace;
    if(permindex!=(int *)NULL)     delete[] permindex;
  }
  vecsize=0;
  Uforward=Uinverse=workspace=(MyComplex *)NULL;
  permindex=(int *)NULL;
}

FFT::~FFT(void)
{
  ReleaseMemory();
}

void FFT::Setup(int size) const
{
  if(size==vecsize) return;
  if(size<1) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),"Error in FFT::Setup(int): "
		"Requested length (%d) must be >0",size);
    OC_THROW(Oxs_BadParameter,buf);
  }
  int k;
  // Check that size is power of 2
  for(k=size;k>2;k/=2) {
    if(k%2!=0) {
      char buf[512];
      Oc_Snprintf(buf,sizeof(buf),"Error in FFT::Setup(int): "
		  "Requested length (%d) is not a power of 2",size);
      OC_THROW(Oxs_BadParameter,buf);
    }
  }
  ReleaseMemory();
  vecsize=size;

  // Allocate and setup MyComplex arrays
  if((Uforward=new MyComplex[size])==0)
    OC_THROW(Oxs_NoMem,"Out of memory error in FFT::Setup");
  if((Uinverse=new MyComplex[size])==0)
    OC_THROW(Oxs_NoMem,"Out of memory error in FFT::Setup");
  if((workspace=new MyComplex[size])==0)
    OC_THROW(Oxs_NoMem,"Out of memory error in FFT::Setup");
#ifdef ORIG_CODE
  REALWIDE baseang= -2*PI/double(size);
  Uforward[0]=Uinverse[0]=MyComplex(1,0);
  REALWIDE x,y;
  for(k=1;k<size/2;k++) {
    x = cos(baseang*k);
    y = -sqrt(1-x*x);
    y += (1-(x*x+y*y))/(2*y);  // Tiny error correction
    x += (1-(x*x+y*y))/(2*x);
    Uforward[k]=Uinverse[size-k]=MyComplex(x,y);
    Uforward[size-k]=Uinverse[k]=MyComplex(x,-y);
  }
  if(size>1) {
    Uforward[size/2]=Uinverse[size/2]=MyComplex(-1,0);
  }
#else
  double baseang = -2*PI/double(size);
  for(k=1;k<size/8;k++) {
    double angle=k*baseang;
    double y=sin(angle);
    double x=cos(angle);

    Uforward[k]=Uinverse[size-k]=MyComplex(x,y);
    Uforward[size/2-k]=Uinverse[size/2+k]=MyComplex(-x,y);
    Uforward[size/2+k]=Uinverse[size/2-k]=MyComplex(-x,-y);
    Uforward[size-k]=Uinverse[k]=MyComplex(x,-y);

    Uforward[size/4-k]=Uinverse[3*size/4+k]=MyComplex(-y,-x);
    Uforward[size/4+k]=Uinverse[3*size/4-k]=MyComplex(y,-x);
    Uforward[3*size/4-k]=Uinverse[size/4+k]=MyComplex(y,x);
    Uforward[3*size/4+k]=Uinverse[size/4-k]=MyComplex(-y,x);
  }
  Uforward[0]=Uinverse[0]=MyComplex(1,0);
  if(size>1) {
    Uforward[size/2]=Uinverse[size/2]=MyComplex(-1,0);
  }
  if(size>3) {
    Uforward[size/4]=Uinverse[3*size/4]=MyComplex(0,-1);
    Uforward[3*size/4]=Uinverse[size/4]=MyComplex(0,1);
  }
  if(size>7) {
    double x=SQRT1_2;  // 1/sqrt(2)
    double y=-x;
    Uforward[size/8]=Uinverse[7*size/8]=MyComplex(x,y);
    Uforward[3*size/8]=Uinverse[5*size/8]=MyComplex(-x,y);
    Uforward[5*size/8]=Uinverse[3*size/8]=MyComplex(-x,-y);
    Uforward[7*size/8]=Uinverse[size/8]=MyComplex(x,-y);
  }
#endif // ORIG_CODE

  // Allocate and setup (bit-reversal) permutation index
  if((permindex=new int[size])==0)
    OC_THROW(Oxs_NoMem,"Out of memory error in FFT::Setup");
  permindex[0]=0;
  int m,n;    // The following code relies heavily on size==2^log2vecsize
  for(k=1,n=size>>1;k<size;k++) {
    // At each step, n is bit-reversed pattern of k
    if(n>k) permindex[k]=n;  // Swap index
    else permindex[k]=0;     // Do nothing: Index already swapped or the same
    // Calculate next n
    m=size>>1;
    while(m>0 && n&m) { n-=m; m>>=1; }
    n+=m;
  }
}

void FFT::ForwardDecFreq(int size,MyComplex *vec,FFT_REAL_TYPE divisor) const
{
  if(divisor==0) divisor=1.0;  // Default is no normalization on forward FFT
  Setup(size);
  BaseDecFreqForward(vec);
  Permute(vec);
  if(divisor!=0. && divisor!=1.) {
    MY_COMPLEX_REAL_TYPE mult=1./divisor;
    for(int k=0;k<size;k++) vec[k]*=mult;
  }
}

void FFT::InverseDecTime(int size,MyComplex *vec,FFT_REAL_TYPE divisor) const
{
  if(divisor==0) divisor=(FFT_REAL_TYPE)size;
  /// Default divisor on iFFT is 'size'
  Setup(size);
  Permute(vec);
  BaseDecTimeInverse(vec);
  if(divisor!=0. && divisor!=1.) {
    MY_COMPLEX_REAL_TYPE mult=1./divisor;
    for(int k=0;k<size;k++) vec[k]*=mult;
  }

}

inline void Swap(MyComplex &a,MyComplex &b)
{ MyComplex c(a); a=b; b=c; }

void FFT::Permute(MyComplex *vec) const
{ /* Bit reversal permutation */
  int i,j;
  for(i=0;i<vecsize;i++) {
    if((j=permindex[i])!=0) Swap(vec[i],vec[j]);
  }
}

void FFT::BaseDecFreqForward(MyComplex *vec) const
{ // In-place forward FFT using Decimation in Frequency technique,
  // *WITHOUT* resuffling of indices.
  // NOTE 1: This code does not use the Complex class, because
  //         some compilers do not effectively optimize around
  //         Complex operations such as multiplication.  So this
  //         routine just makes use of ordinary type "FFT_REAL_TYPE"
  //         variables, and assumes each Complex variable is
  //         actually two consecutive "MY_COMPLEX_REAL_TYPE" variables.
  // NOTE 2: See notes in MJD's micromagnetics notebook, 11-Sep-96
  //         (p. 62) and 29-Sep-96 (p. 69).
  // NOTE 3: This code has been optimized for performance on cascade.cam,
  //         a PentiumPro 200 MHz machine using a stock gcc 2.7.2
  //         compiler.  In particular, the x86 chips suffer from a
  //         shortage of registers.
  // NOTE 4: Some compromise made to RISC architectures on 27-May-1997,
  //         by moving all loads before any stores in main loop.  As
  //         done, it hurts performance on PentiumPro-200 only a couple
  //         of percent. (mjd)

  if(vecsize==1) return; // Nothing to do

  MY_COMPLEX_REAL_TYPE *v;

  MY_COMPLEX_REAL_TYPE const *const U=(MY_COMPLEX_REAL_TYPE *)Uforward;
  MY_COMPLEX_REAL_TYPE *const dvec=(MY_COMPLEX_REAL_TYPE *)vec;

  int block,blocksize,blockcount,offset,uoff1;
  int halfbs,threehalfbs; // Half blocksize,3/2 blocksize
  FFT_REAL_TYPE m1x,m1y,m2x,m2y,m3x,m3y;
  FFT_REAL_TYPE x0,y0,x1,y1,x2,y2,x3,y3;
  FFT_REAL_TYPE xs02,ys02,xd02,yd02,xs13,ys13,xd13,yd13;
  FFT_REAL_TYPE t1x,t1y,t2x,t2y,t3x,t3y;


  // Blocksize>4
  for(blocksize=vecsize,blockcount=1;blocksize>4;
      blocksize/=4,blockcount*=4) {
    // Loop through double-step matrix multiplications
    halfbs=blocksize/2; threehalfbs=blocksize+halfbs;
    for(block=0,v=dvec;block<blockcount;block++,v+=2*blocksize) {
      for(offset=0;offset<halfbs;offset+=2) {
	uoff1=offset*blockcount;
	m1x=U[uoff1];	m1y=U[uoff1+1];
	m2x=U[2*uoff1];	m2y=U[2*uoff1+1];
	m3x=U[3*uoff1];	m3y=U[3*uoff1+1];

	x0=v[offset];
        y0=v[offset+1];
	x1=v[halfbs+offset];
	y1=v[halfbs+offset+1];
	x2=v[blocksize+offset];
	y2=v[blocksize+offset+1];
	x3=v[threehalfbs+offset];
	y3=v[threehalfbs+offset+1];

	xs02=x0+x2;	xs13=x1+x3;
	v[offset]=xs02+xs13;
	t1x=xs02-xs13;

	ys02=y0+y2;	ys13=y1+y3;
        v[offset+1]=ys02+ys13;
	t1y=ys02-ys13;

	v[halfbs+offset]    =t1x*m2x-t1y*m2y;
	v[halfbs+offset+1]  =t1y*m2x+t1x*m2y;

	xd02=x0-x2;	yd13=y1-y3;
	t3x=xd02-yd13;  t2x=xd02+yd13;
        yd02=y0-y2;	xd13=x1-x3;
	t3y=yd02+xd13;  t2y=yd02-xd13;
	v[blocksize+offset]  =t2x*m1x-t2y*m1y;
	v[blocksize+offset+1]=t2y*m1x+t2x*m1y;
	v[threehalfbs+offset]  =t3x*m3x-t3y*m3y;
	v[threehalfbs+offset+1]=t3y*m3x+t3x*m3y;
      }
    }
  }

  // Do smallest blocks; size is either 4 or 2
  if(blocksize==4) {
    blockcount=vecsize/4;
    for(block=0,v=dvec;block<blockcount;block++,v+=8) {
      x0=v[0];      y0=v[1];      x1=v[2];      y1=v[3];
      x2=v[4];      y2=v[5];      x3=v[6];      y3=v[7];

      xs02=x0+x2;
      xs13=x1+x3; 
      v[0]=xs02+xs13;
      v[2]=xs02-xs13;

      ys02=y0+y2;
      ys13=y1+y3;
      v[1]=ys02+ys13;
      v[3]=ys02-ys13;

      xd02=x0-x2;
      yd13=y1-y3;
      v[4]=xd02+yd13;
      v[6]=xd02-yd13;

      yd02=y0-y2;
      xd13=x1-x3;
      v[5]=yd02-xd13;
      v[7]=yd02+xd13;
    }
  }
  else { // blocksize==2
    blockcount=vecsize/2;
    for(block=0,v=dvec;block<blockcount;block++,v+=4) {
      x0=v[0];      y0=v[1];
      x1=v[2];      y1=v[3];
      v[0]=x0+x1;   v[2]=x0-x1;
      v[1]=y0+y1;   v[3]=y0-y1;
    }
  }
}

void FFT::BaseDecTimeInverse(MyComplex *vec) const
{ // In-place inverse FFT using Decimation in Time technique,
  // *WITHOUT* resuffling of indices.
  // NOTE 1: This code does not use the Complex class, because
  //         some compilers do not effectively optimize around
  //         Complex operations such as multiplication.  So this
  //         routine just makes use of ordinary type "FFT_REAL_TYPE"
  //         variables, and assumes each Complex variable is
  //         actually two consecutive "MY_COMPLEX_REAL_TYPE" variables.
  // NOTE 2: See notes in MJD's micromagnetics notebook, 11-Sep-96
  //         (p. 62) and 29-Sep-96 (p. 69).
  // NOTE 3: This code has been optimized for performance on cascade.cam,
  //         a PentiumPro 200 MHz machine using a stock gcc 2.7.2
  //         compiler.  In particular, the x86 chips suffer from a
  //         shortage of registers.
  // NOTE 4: Some compromise made to RISC architectures on 27-May-1997,
  //         by moving all loads before any stores in main loop.  As
  //         done, it hurts performance on PentiumPro-200 only a couple
  //         of percent. (mjd)

  if(vecsize==1) return; // Nothing to do

  MY_COMPLEX_REAL_TYPE *v;

  MY_COMPLEX_REAL_TYPE const *const U=(MY_COMPLEX_REAL_TYPE *)Uinverse;
  MY_COMPLEX_REAL_TYPE *const dvec=(MY_COMPLEX_REAL_TYPE *)vec;

  int block,blocksize,blockcount,offset,uoff1;
  int halfbs,threehalfbs; // Half blocksize,3/2 blocksize
  FFT_REAL_TYPE m1x,m1y,m2x,m2y,m3x,m3y;
  FFT_REAL_TYPE x0,y0,x1,y1,x2,y2,x3,y3;
  FFT_REAL_TYPE xs01,ys01,xd01,yd01,xs23,ys23,xd23,yd23;
  FFT_REAL_TYPE t1x,t1y,t2x,t2y,t3x,t3y;

  // Do smallest blocks; size is either 4 or 2
  if(vecsize>2) {
    blockcount=vecsize/4;
    for(block=0,v=dvec;block<blockcount;block++,v+=8) {
      x0=v[0];      y0=v[1];
      x1=v[2];      y1=v[3];
      x2=v[4];      y2=v[5];
      x3=v[6];      y3=v[7];
      xs01=x0+x1;      xs23=x2+x3;   // See NOTE 3 above
      v[0]=xs01+xs23;      v[4]=xs01-xs23;
      ys01=y0+y1;      ys23=y2+y3;
      v[1]=ys01+ys23;      v[5]=ys01-ys23;
      xd01=x0-x1;      yd23=y2-y3;
      v[2]=xd01-yd23;      v[6]=xd01+yd23;
      yd01=y0-y1;      xd23=x2-x3;
      v[3]=yd01+xd23;      v[7]=yd01-xd23;
    }
  }
  else { // vecsize==2
    x0=dvec[0];    y0=dvec[1];
    x1=dvec[2];    y1=dvec[3];
    dvec[0]=x0+x1;
    dvec[2]=x0-x1;
    dvec[1]=y0+y1;
    dvec[3]=y0-y1;
    return;
  }

  // Blocksize>4
  for(blocksize=16,blockcount=vecsize/16;blocksize<=vecsize;
      blocksize*=4,blockcount/=4) {
    // Loop through double-step matric multiplications
    halfbs=blocksize/2; threehalfbs=blocksize+halfbs;
    for(block=0,v=dvec;block<blockcount;block++,v+=2*blocksize) {
      for(offset=0;offset<blocksize/2;offset+=2) {
	x0=v[offset];
	y0=v[offset+1];
	t2x=v[blocksize+offset];
	t2y=v[blocksize+offset+1];
	uoff1=offset*blockcount;
	m1x=U[uoff1];  m1y=U[uoff1+1];
	x2=t2x*m1x-t2y*m1y;
	y2=t2y*m1x+t2x*m1y;

	m2x=U[2*uoff1];  m2y=U[2*uoff1+1];
	t1x=v[halfbs+offset];
	t1y=v[halfbs+offset+1];
	x1=t1x*m2x-t1y*m2y;
	y1=t1y*m2x+t1x*m2y;

	t3x=v[threehalfbs+offset];
	t3y=v[threehalfbs+offset+1];
	m3x=U[3*uoff1];  m3y=U[3*uoff1+1];
	x3=t3x*m3x-t3y*m3y;
	y3=t3y*m3x+t3x*m3y;

	xs01=x0+x1;	xs23=x2+x3;
	v[            offset  ] = xs01+xs23;
	v[  blocksize+offset  ] = xs01-xs23;

	ys01=y0+y1;     ys23=y2+y3;
        v[            offset+1] = ys01+ys23;
	v[  blocksize+offset+1] = ys01-ys23;

	yd01=y0-y1;	xd23=x2-x3;
	v[  halfbs   +offset+1] = yd01+xd23;
	v[threehalfbs+offset+1] = yd01-xd23;

	xd01=x0-x1;	yd23=y2-y3;
	v[  halfbs   +offset  ] = xd01-yd23;
	v[threehalfbs+offset  ] = xd01+yd23;
      }
    }
  }

  if(blocksize==2*vecsize) {
    // We still have to do one single-step matrix multiplication
    blocksize=vecsize;  v=dvec;
    for(offset=0;offset<blocksize;offset+=2,v+=2) {
      x0=v[0];              y0=v[1];
      t1x=v[vecsize];       t1y=v[vecsize+1];
      m1x=U[offset];        m1y=U[offset+1];
      x1=t1x*m1x-t1y*m1y;   y1=t1y*m1x+t1x*m1y;
      v[0]         = x0+x1; 
      v[vecsize]   = x0-x1;
      v[1]         = y0+y1;
      v[vecsize+1] = y0-y1;
    }
  }

}

////////////////////////////////////////////////////////////////////////
// 3D Complex FFT
const int FFT3D::scratch_block_size=4;  // Number of lines to block

void FFT3D::ReleaseMemory() const
{
  if(scratchsize>0 && scratch!=NULL) delete[] scratch;
  scratchsize=0;
  scratch=NULL;
  xfft.ReleaseMemory();
  yfft.ReleaseMemory();
  zfft.ReleaseMemory();
}

void FFT3D::AdjustScratchSize(UINT4m xsize,UINT4m ysize,UINT4m zsize) const
{
  UINT4m maxsize=xsize;
  if(ysize>maxsize) maxsize=ysize;
  if(zsize>maxsize) maxsize=zsize;
  if(scratchsize<maxsize) {
    if(scratchsize>0 && scratch!=NULL) delete[] scratch;
    scratchsize=maxsize;
    scratch=new MyComplex[scratch_block_size*scratchsize];
  }
}


void
FFT3D::Forward(UINT4m xsize,UINT4m ysize,UINT4m zsize,
	       MyComplex* arr) const
{ // Conceptually const
  UINT4m totalsize=xsize*ysize*zsize;
  if(xsize==0 || ysize==0 || zsize==0) return; // Nothing to do
  if(totalsize < xsize || totalsize < ysize || totalsize < zsize) {
    // Partial overflow check
    string msg="UINT4m overflow in FFT3D::Forward(): "
      "Product xsize*ysize*zsize too big to fit in a UINT4m variable";
    OC_THROW(Oxs_Overflow,msg.c_str());
  }

  AdjustScratchSize(xsize,ysize,zsize);

  UINT4m i,j,k,block,block_stop;
  UINT4m offset;
  UINT4m xysize = xsize*ysize;
  MyComplex* aptr;

  // Do 1D FFT's in x-direction
  aptr = arr;
  for(k=0;k<zsize;k++) for(j=0;j<ysize;j++) {
      xfft.ForwardDecFreq(xsize,aptr);
      aptr += xsize;
  }
  
  // Do 1D FFT's in y-direction
  for(k=0;k<zsize;k++) {
    aptr = arr + k*xysize;  // Set base z-position
    for(i=0;i<xsize;i+=scratch_block_size) {
      if(i+scratch_block_size<=xsize) block_stop=scratch_block_size;
      else                            block_stop=xsize-i;
      offset=0;
      for(j=0;j<ysize;j++) {
	int j2;
	for(block=0,j2=j;block<block_stop;block++,j2+=ysize) {
	  // Copy from arr into scratch buffer
	  scratch[j2] = aptr[offset+block];
	}
	offset+=xsize;
      }
      for(block=0;block<block_stop;block++) {
	// Do FFT in scratch buffer
	yfft.ForwardDecFreq(ysize,scratch+block*ysize);
      }
      offset=0;
      for(j=0;j<ysize;j++) {
	int j2;
	for(block=0,j2=j;block<block_stop;block++,j2+=ysize) {
	  // Copy from arr into scratch buffer
	  aptr[offset+block] = scratch[j2];
	}
	offset+=xsize;
      }
      aptr+=block_stop; // Increment base x-position
    }
  }

  // Do 1D FFT's in z-direction
  aptr=arr;
  for(j=0;j<ysize;j++) {
    for(i=0;i<xsize;i+=scratch_block_size) {
      if(i+scratch_block_size<=xsize) block_stop=scratch_block_size;
      else                            block_stop=xsize-i;
      offset=0;
      for(k=0;k<zsize;k++) {
	int k2;
	for(block=0,k2=k;block<block_stop;block++,k2+=zsize) {
	  // Copy from arr into scratch buffer
	  scratch[k2] = aptr[offset+block];
	}
	offset+=xysize;
      }
      for(block=0;block<block_stop;block++) {
	// Do FFT in scratch buffer
	zfft.ForwardDecFreq(zsize,scratch+block*zsize);
      }
      offset=0;
      for(k=0;k<zsize;k++) {
	int k2;
	for(block=0,k2=k;block<block_stop;block++,k2+=zsize) {
	  // Copy from arr into scratch buffer
	  aptr[offset+block] = scratch[k2];
	}
	offset+=xysize;
      }
      aptr+=block_stop; // Increment base position
    }
  }

}

void
FFT3D::Inverse(UINT4m xsize,UINT4m ysize,UINT4m zsize,
	       MyComplex* arr) const
{ // Conceptually const
  UINT4m totalsize=xsize*ysize*zsize;
  if(xsize==0 || ysize==0 || zsize==0) return; // Nothing to do
  if(totalsize < xsize || totalsize < ysize || totalsize < zsize) {
    // Partial overflow check
    string msg="UINT4m overflow in FFT3D::Inverse(): "
      "Product xsize*ysize*zsize too big to fit in a UINT4m variable";
    OC_THROW(Oxs_Overflow,msg.c_str());
  }

  AdjustScratchSize(xsize,ysize,zsize);

  UINT4m i,j,k,block,block_stop;
  UINT4m offset;
  UINT4m xysize = xsize*ysize;
  MyComplex* aptr;

  // Do 1D iFFT's in z-direction
  aptr=arr;
  for(j=0;j<ysize;j++) {
    for(i=0;i<xsize;i+=scratch_block_size) {
      if(i+scratch_block_size<=xsize) block_stop=scratch_block_size;
      else                            block_stop=xsize-i;
      offset=0;
      for(k=0;k<zsize;k++) {
	int k2;
	for(block=0,k2=k;block<block_stop;block++,k2+=zsize) {
	  // Copy from arr into scratch buffer
	  scratch[k2] = aptr[offset+block];
	}
	offset+=xysize;
      }
      for(block=0;block<block_stop;block++) {
	// Do iFFT in scratch buffer
	zfft.InverseDecTime(zsize,scratch+block*zsize,1.);
      }
      offset=0;
      for(k=0;k<zsize;k++) {
	int k2;
	for(block=0,k2=k;block<block_stop;block++,k2+=zsize) {
	  // Copy from arr into scratch buffer
	  aptr[offset+block] = scratch[k2];
	}
	offset+=xysize;
      }
      aptr+=block_stop; // Increment base position
    }
  }
  
  // Do 1D iFFT's in y-direction
  for(k=0;k<zsize;k++) {
    aptr = arr + k*xysize;  // Set base z-position
    for(i=0;i<xsize;i+=scratch_block_size) {
      if(i+scratch_block_size<=xsize) block_stop=scratch_block_size;
      else                            block_stop=xsize-i;
      offset=0;
      for(j=0;j<ysize;j++) {
	int j2;
	for(block=0,j2=j;block<block_stop;block++,j2+=ysize) {
	  // Copy from arr into scratch buffer
	  scratch[j2] = aptr[offset+block];
	}
	offset+=xsize;
      }
      for(block=0;block<block_stop;block++) {
	// Do FFT in scratch buffer
	yfft.InverseDecTime(ysize,scratch+block*ysize,1.);
      }
      offset=0;
      for(j=0;j<ysize;j++) {
	int j2;
	for(block=0,j2=j;block<block_stop;block++,j2+=ysize) {
	  // Copy from arr into scratch buffer
	  aptr[offset+block] = scratch[j2];
	}
	offset+=xsize;
      }
      aptr+=block_stop; // Increment base x-position
    }
  }

  // Do 1D iFFT's in x-direction
  aptr = arr;
  for(k=0;k<zsize;k++) for(j=0;j<ysize;j++) {
      xfft.InverseDecTime(xsize,aptr,1.);
      aptr += xsize;
  }

  // Scale
  REAL8m scale = 1./static_cast<REAL8m>(totalsize);
  for(i=0;i<totalsize;i++) arr[i] *= scale;

}
