/* FILE: simpledemag.h            -*-Mode: c++-*-
 *
 * Average H demag field across rectangular cells.  Simple
 * implementation that does not take advantage of any memory
 * saving or calculation speed improvements possible using
 * symmetries in interaction tensor or realness of data in
 * spatial domain.
 *
 */

#include "nb.h"
#include "director.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

#include "rectangularmesh.h"
#include "simpledemag.h"
#include "demagcoef.h"
#include "fft.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_SimpleDemag);

/* End includes */

// Helper function
UINT4m Oxs_SimpleDemag::NextPowerOfTwo(UINT4m n) const
{ // Returns first power of two >= n
  UINT4m m=1;
  while(m<n) m*=2;
  return m;
}

// Constructor
Oxs_SimpleDemag::Oxs_SimpleDemag(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr),
    xdim(0),ydim(0),zdim(0),totalsize(0),
    pxdim(0),pydim(0),pzdim(0),ptotalsize(0),
    A00(NULL),A01(NULL),A02(NULL),A11(NULL),A12(NULL),A22(NULL),
    Mx(NULL),My(NULL),Mz(NULL),Hcomp(NULL)
{

  CheckInitValueParamCount("mesh",1);
  const string& meshname = (*FindInitValue("mesh"))[0];
  Oxs_Ext* tempobj = director->FindExtObject(meshname);
  if(tempobj==NULL) {
    string msg="Unable to identify Oxs_Mesh object with name "
      + meshname;
    throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_RectangularMesh* mesh
    = dynamic_cast<Oxs_RectangularMesh*>(tempobj);
  if(mesh==NULL) {
    string msg="Object " + meshname + " is not a rectangular mesh.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  meshkey.Set(mesh); // Dependency lock is held until *this is deleted.
  DeleteInitValue("mesh");

  VerifyAllInitArgsUsed();

  // Fill dimension variables
  xdim = mesh->DimX();
  ydim = mesh->DimY();
  zdim = mesh->DimZ();
  totalsize=xdim*ydim*zdim;
  if(xdim==0 || ydim==0 || zdim==0) return; // Empty mesh!
  if(totalsize < xdim || totalsize < ydim || totalsize < zdim) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product xdim*ydim*zdim too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }
  pxdim = NextPowerOfTwo(2*xdim);
  pydim = NextPowerOfTwo(2*ydim);
  pzdim = NextPowerOfTwo(2*zdim);
  ptotalsize=pxdim*pydim*pzdim;
  if(ptotalsize < pxdim || ptotalsize < pydim || ptotalsize < pzdim) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product pxdim*pydim*pzdim too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }


  // Allocate memory for interaction matrices and magnetization components
  A00   = new MyComplex[ptotalsize];
  A01   = new MyComplex[ptotalsize];
  A02   = new MyComplex[ptotalsize];
  A11   = new MyComplex[ptotalsize];
  A12   = new MyComplex[ptotalsize];
  A22   = new MyComplex[ptotalsize];
  Mx    = new MyComplex[ptotalsize];
  My    = new MyComplex[ptotalsize];
  Mz    = new MyComplex[ptotalsize];
  Hcomp = new MyComplex[ptotalsize];
  if(Hcomp==NULL || Mx==NULL || My==NULL || Mz==NULL
     || A00==NULL || A01==NULL || A02==NULL
     || A11==NULL || A12==NULL || A22==NULL) {
    // Safety check for those machines on which new[] doesn't throw
    // BadAlloc.
    string msg="Insufficient memory in simpledemag constructor.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Initialize interaction matrices to zeros
  UINT4m pindex;
  for(pindex=0;pindex<ptotalsize;pindex++) A00[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A01[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A02[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A11[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A12[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A22[pindex].Set(0.,0.);

  // According (16) in Newell's paper, the demag field is given by
  //                        H = -N*M
  // where N is the "demagnetizing tensor," with components Nxx, Nxy,
  // etc.  With the '-1' in 'scale' we store '-N' instead of 'N',
  // so we don't have to multiply the output from the FFT + iFFT
  // by -1 in ConstMagField() below.

  // Fill interaction matrices with demag coefs from Newell's paper.
  // Note that A00, A11 and A22 are even in x,y and z.
  // A01 is odd in x and y, even in z.
  // A02 is odd in x and z, even in y.
  // A12 is odd in y and z, even in x.
  UINT4m i,j,k;
  UINT4m pxydim=pxdim*pydim;
  REALWIDE dx = mesh->EdgeLengthX();
  REALWIDE dy = mesh->EdgeLengthY();
  REALWIDE dz = mesh->EdgeLengthZ();
  // For demag calculation, all that matters is the relative
  // sizes of dx, dy and dz.  To help insure we don't run
  // outside floating point range, rescale these values so
  // largest is 1.0
  REALWIDE maxedge=dx;
  if(dy>maxedge) maxedge=dy;
  if(dz>maxedge) maxedge=dz;
  dx/=maxedge; dy/=maxedge; dz/=maxedge;
  REALWIDE scale = -1./(4*PI*dx*dy*dz);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE x = dx*i;
    REALWIDE y = dy*j;
    REALWIDE z = dz*k;
    REALWIDE coef=scale*CalculateSDA00(x,y,z,dx,dy,dz);
    A00[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A00[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A00[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A00[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A00[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A00[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A00[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A00[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE x = dx*i;
    REALWIDE y = dy*j;
    REALWIDE z = dz*k;
    REALWIDE coef=scale*CalculateSDA11(x,y,z,dx,dy,dz);
    A11[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A11[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A11[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A11[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A11[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A11[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A11[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A11[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE x = dx*i;
    REALWIDE y = dy*j;
    REALWIDE z = dz*k;
    REALWIDE coef=scale*CalculateSDA22(x,y,z,dx,dy,dz);
    A22[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A22[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A22[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A22[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A22[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A22[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A22[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A22[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE x = dx*i;
    REALWIDE y = dy*j;
    REALWIDE z = dz*k;
    REALWIDE coef=scale*CalculateSDA01(x,y,z,dx,dy,dz);
    A01[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A01[(pxdim-i)+j*pxdim+k*pxydim].Set(-coef,0.);
    if(j>0) A01[i+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(k>0) A01[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A01[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A01[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(j>0 && k>0)
      A01[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0 && k>0)
      A01[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE x = dx*i;
    REALWIDE y = dy*j;
    REALWIDE z = dz*k;
    REALWIDE coef=scale*CalculateSDA02(x,y,z,dx,dy,dz);
    A02[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A02[(pxdim-i)+j*pxdim+k*pxydim].Set(-coef,0.);
    if(j>0) A02[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A02[i+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0)
      A02[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(i>0 && k>0)
      A02[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A02[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0 && k>0)
      A02[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE x = dx*i;
    REALWIDE y = dy*j;
    REALWIDE z = dz*k;
    REALWIDE coef=scale*CalculateSDA12(x,y,z,dx,dy,dz);
    A12[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A12[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A12[i+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(k>0) A12[i+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0)
      A12[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(i>0 && k>0)
      A12[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(j>0 && k>0)
      A12[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A12[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }

  // Transform interaction matrices into frequency domain.
  fft.Forward(pxdim,pydim,pzdim,A00);
  fft.Forward(pxdim,pydim,pzdim,A01);
  fft.Forward(pxdim,pydim,pzdim,A02);
  fft.Forward(pxdim,pydim,pzdim,A11);
  fft.Forward(pxdim,pydim,pzdim,A12);
  fft.Forward(pxdim,pydim,pzdim,A22);
}

void Oxs_SimpleDemag::ReleaseMemory()
{
  if(Mx!=NULL) { delete[] Mx; Mx=NULL; }
  if(My!=NULL) { delete[] My; My=NULL; }
  if(Mz!=NULL) { delete[] Mz; Mz=NULL; }
  if(A00!=NULL) { delete[] A00; A00=NULL; }
  if(A01!=NULL) { delete[] A01; A01=NULL; }
  if(A02!=NULL) { delete[] A02; A02=NULL; }
  if(A11!=NULL) { delete[] A11; A11=NULL; }
  if(A12!=NULL) { delete[] A12; A12=NULL; }
  if(A22!=NULL) { delete[] A22; A22=NULL; }
  xdim=ydim=zdim=totalsize=0;
  pxdim=pydim=pzdim=ptotalsize=0;
}

BOOL Oxs_SimpleDemag::Init()
{
  return 1;
}

void Oxs_SimpleDemag::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  // Check that the mesh hasn't changed.
  if(meshkey.KeyId() != state.mesh->Id()) {
    string msg="Input mesh to Oxs_SimpleDemag::GetEnergyAndField()"
      " routine of object " +  string(InstanceName())
      + " is different than the one used to initialize this object.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);

  // Calculate FFT of Mx, My and Mz
  UINT4m xydim=xdim*ydim;
  UINT4m pxydim=pxdim*pydim;
  UINT4m index,pindex;
  UINT4m i,j,k;
  for(pindex=0;pindex<ptotalsize;pindex++) Mx[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) My[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) Mz[pindex].Set(0.,0.);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    REAL8m scale = Ms[index];
    const ThreeVector& vec = spin[index];
    Mx[pindex].Set(scale*(vec.x),0);
    My[pindex].Set(scale*(vec.y),0);
    Mz[pindex].Set(scale*(vec.z),0);
  }
  fft.Forward(pxdim,pydim,pzdim,Mx);
  fft.Forward(pxdim,pydim,pzdim,My);
  fft.Forward(pxdim,pydim,pzdim,Mz);
  
  // Calculate field components in frequency domain, then to iFFT
  // to transform back to space domain.

  // Hx component
  for(pindex=0;pindex<ptotalsize;pindex++) {
    Hcomp[pindex] = A00[pindex]*Mx[pindex]
      + A01[pindex]*My[pindex] + A02[pindex]*Mz[pindex];
  }
  fft.Inverse(pxdim,pydim,pzdim,Hcomp);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    field[index].x=Hcomp[pindex].real();
  }
  
  // Hy component
  for(pindex=0;pindex<ptotalsize;pindex++) {
    Hcomp[pindex] = A01[pindex]*Mx[pindex]
      + A11[pindex]*My[pindex] + A12[pindex]*Mz[pindex];
  }
  fft.Inverse(pxdim,pydim,pzdim,Hcomp);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    field[index].y=Hcomp[pindex].real();
  }
  
  // Hz component
  for(pindex=0;pindex<ptotalsize;pindex++) {
    Hcomp[pindex] = A02[pindex]*Mx[pindex]
      + A12[pindex]*My[pindex] + A22[pindex]*Mz[pindex];
  }
  fft.Inverse(pxdim,pydim,pzdim,Hcomp);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    field[index].z=Hcomp[pindex].real();
  }

  // Calculate pointwise energy density: -0.5*MU0*<M,H>
  REAL8m mult = -0.5 * MU0;
  for(index=0;index<totalsize;index++) {
    energy[index] = mult * Ms[index] * (spin[index]*field[index]);
  }

}
